//OrderDistiller.java modified by Pietro Terna (look below at rows signed //pp)
//OrderDistiller.java modified by Marco Lamieri and Francesco Merlo 
//(look below at rows signed //lm)

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;

/**
 * OrderDistiller.java
 *
 *
 * Created: Wed May 08 15:29:12 2002
 *
 * @author<br/> 
 * Cristian Barreca, Elena Bonessa and Antonella Borra</br>
 * Modified by:</br>
 * Marco Lamieri and Francesco Merlo
 * @version 0.9.7.31.b
 */

/**
 * This class is used to read data from two worksheets. <br/>The first one contains the 
 * list of recipes of 
 * our virtual enterprise.<br/>The second one contains a sequence of orders to be launched, 
 * shift by shift, in order to make the daily production activities.*/     

public class OrderDistiller extends OrderGenerator{
    
    /** INSTANCE VARIABLES
     *A flag to check if the orderSequence worksheet file is open
     */
    public boolean worksheetOrderSequenceFileOpen = false, worksheetRecipeFileOpen = false;
    
    /**The arrays containing: the sequence of orders to be done in the related shift,
     * the respective quantity and layer
     */ 

    //lm
    //*****************************************************
    // Added third array containing layers info
    //*****************************************************    
    int[] orderSequence1, orderSequence2, orderSequence3;
    
    /**The variable referring to the orderSequence worksheet file
     */
    ExcelReader orderSequenceWorksheet = null, recipeWorksheet = null;
    /** The name of the recipe
     */
    String recipeName;    

    /**Flags to operate checks while reading 
     */

    //lm
    //*****************************************************
    // Added String "layer" and "computation"
    //*****************************************************
    public String semicolon = ";", checkTheCell, gate = "#", p = "p", 
    	sec = "s", min = "m", end = "e", slash = "/", backslash = "\\",
    	or = "||", layer="l", computation="c";
    
    /**An object to record the array containing the steps of the recipes in a List
     */
    Recipe aRecipe;
    
    /** used to record the number of generated orders
     */
    public int orderCount = 0;
    
    //lm
    //*****************************************************
    // used to decide if read "orderStartingSequence.xls" (true) or 
    // "orderSequence.xls" (false)  using readOrderSequence method.
    // It is useful for compatibility with Vir application.
    //*****************************************************
    public static boolean firstTime = true;
       
    
    /** a specific order
     */
    public Order anOrder;
    
    /** the list containig the operating units
     */
    public ListImpl unitList;
    
    /** the list containig the end units
     */
    public ListImpl endUnitList;
    
    /** the list containig all the orders
     */
    public ListImpl orderList;
    
    /** the list containig all the orders
     */
    public ListImpl recipeList;
    
    /**The array containing the steps of the recipe and the steps of procurement
     */
    int[] orderRecipe;
    
    // units
    Unit aUnit;
    EndUnit anEndUnit;
    
    /* public boolean unitNotFound;*/ //pt
    int j, row, length, choice;

	//lm
	//*****************************************************
	// Added variable for current layer 
	//*****************************************************
     	
	int currentLayer = 0;


	//lm
	//*****************************************************
	// Added variable for orderSequenceX dimension
	//*****************************************************
	
	int maxOrderSequence; 

    AssigningTool assigningTool; //pt
    




    // CONSTRUCTOR
    public OrderDistiller (Zone aZone, int msn, int msl, ListImpl ul, 
			   ListImpl eul, ListImpl ol, int tln, 
			   ESFrameModelSwarm mo, AssigningTool at){ //pt
	
	super(aZone, msn, msl, ul, eul, ol, tln, mo, at);
	
	unitList=ul;
	endUnitList=eul;
        orderList=ol;
	assigningTool=at; //pt
	
    }

   
    /**This method is used to collect the names of the units and of the end units, so that it can operate the check of 
     * corrispondency between the production phases required by recipes and the phases of production the units can do. 
     */
    public void setDictionary(){
	
	super.setDictionary();
		
    //*****************************************************
    // Added "orderSequence3" array containing layers info //lm
    //****************************************************
    
      
    maxOrderSequence = dictionaryLength * totalLayerNumber;    
    
    // It will contain the ID codes of the recipes worksheet file  
	orderSequence1 = new int[maxOrderSequence];
	// It will contain the quantity of each recipe to be done  
	orderSequence2 = new int[maxOrderSequence];
	// It will contain the layer of each recipe to be done  
	orderSequence3 = new int[maxOrderSequence];
		
	setRecipeContainers(); //See Below
	readRecipes(); //See Below
    }
    
    /**This method is used to set the length of each recipe
     */
    public void setRecipeContainers(){
	
	recipeList = new ListImpl(getZone());
	
	if(! worksheetRecipeFileOpen){ 
	    recipeWorksheet = new ExcelReader("recipeData/recipes.xls");
	    worksheetRecipeFileOpen = true;
	}
	
	checkTheCell = recipeWorksheet.getStrValue();
	
	row = 0;
	while(! recipeWorksheet.eof()){	
	    checkForComments(checkTheCell);
	    calculateLength(checkTheCell);
	}
	
	worksheetRecipeFileOpen = false;		
    }
    
    /** This is the method needed to read and store the recipes. The procedure follows this steps:<br/> It opens the worksheet file recipes.xls, in which are stored the sequence of steps of all the recipes;<br/>It makes some check of the routines; <br/>It search the object containig the same code of the recipe we are considering;<br/>
     * Finally it substitutes the values of the array of that object with the new ones.  
     */
    public void readRecipes(){
	//local variables
	int code = 0, i;
	boolean recipeCodeNotFound;
	
	if(! worksheetRecipeFileOpen){ 
	    recipeWorksheet = new ExcelReader("recipeData/recipes.xls");
	    worksheetRecipeFileOpen = true;
	}
	row = 0;
	checkTheCell = recipeWorksheet.getStrValue();
	while(! recipeWorksheet.eof()){	
	    
	    checkForComments(checkTheCell);
	    code = errorIsNotAnInteger(recipeWorksheet);
	    
	    recipeCodeNotFound=true;
	    for (i = 0; i < recipeList.getCount() && recipeCodeNotFound; i++)
		{
		    aRecipe = (Recipe) recipeList.atOffset(i);
		    if(aRecipe.getCodeNumber() == code)recipeCodeNotFound = false;
		}
	    
	    aRecipe.setSteps(checkTheCell, recipeWorksheet);
	    checkTheCell = aRecipe.getCheckTheCell();
	}
    }
       
    /**This method is used to calculate the length of each row after a strong check of the elements
     */
    public void calculateLength(String cTC){
	int code = 0;
	checkTheCell = cTC;
	recipeName = checkTheCell;	
	code = errorIsNotAnInteger(recipeWorksheet);
	row++;
	length = 0;
	checkTheCell = recipeWorksheet.getStrValue();
	
	
	while(! checkTheCell.equals(semicolon)){
	    if(StartESFrame.verbose)
		System.out.println("CheckTheCell contains "+checkTheCell);


	    if(checkTheCell.equals(p))   			choice = 1;
	    if(checkTheCell.equals(or))  			choice = 2;
	    if(checkTheCell.equals(end)) 			choice = 3;
	    if(checkTheCell.equals(computation)) 	choice = 4;
	    if(recipeWorksheet.checkForLabelCell() && !checkTheCell.equals(semicolon))   choice = 5;
	  
	    switch(choice){
	    case 1:
			if(StartESFrame.verbose)
			System.out.println("OrderDistiller: the choice is for a Procurement");
			procurement(recipeWorksheet);
		break;
	    case 2:
			if(StartESFrame.verbose)
			System.out.println("OrderDistiller: the choice is for an Or ");
			oR(recipeWorksheet);
		break;
	    case 3:
			if(StartESFrame.verbose)
			System.out.println("OrderDistiller: the choice is for an End ");
			end(recipeWorksheet);
		break;
    	case 4:
			if(StartESFrame.verbose)
			System.out.println("OrderDistiller: the choice is for a Computation ");
			computation(recipeWorksheet);
		break;			
	    case 5:
			if(StartESFrame.verbose)
			System.out.println("OrderDistiller: the choice is for a Number");
			number(recipeWorksheet);
		break;
	    default:
	if(StartESFrame.verbose)
		System.out.println("OrderDistiller: no matches were found reading the worksheet, check for errors inside it");
		System.exit(1);


	    }
	}
	
	if(! recipeWorksheet.eof())
	    checkTheCell = recipeWorksheet.getStrValue();

	
	
	aRecipe = new Recipe(getZone(), code, length, recipeName);
	if( StartESFrame.verbose)
	    System.out.println("OrderDistiller: a Recipe named "+ recipeName+ " with code " + code + "was born");
	recipeList.addLast(aRecipe);
    }


 	//lm
	// ****************************************
	// A new method for the computational step 
	// ****************************************   
    
     /**This method dial with the computational choice
     */    
    public void computation(ExcelReader e){
	int numberOfMatrixesForComputation = 0;
	int computationCode;
	
	// The computational code	
	e.getIntValue();

	numberOfMatrixesForComputation = e.getIntValue();
	length += (3 + numberOfMatrixesForComputation);
	
	// Skipping the Matrixes
	for(int h = 0; h < numberOfMatrixesForComputation; h++) checkTheCell = e.getStrValue();
	checkTheCell = e.getStrValue();

    } 
    
    /**This method dial with the procurement choice
     */    
    public void procurement(ExcelReader e){
	int numberOfStepsForProcurement = 0;

	numberOfStepsForProcurement = e.getIntValue();
	length += (2 + numberOfStepsForProcurement);
	for(int h = 0; h < numberOfStepsForProcurement; h++) checkTheCell = e.getStrValue();
	checkTheCell = e.getStrValue();

    } 
    

    /**This method dial with the or choice
     */
    public void oR(ExcelReader e){

	length +=2;
	e.getStrValue();
	checkTheCell = e.getStrValue();

    }

    /**This method dial with the end choice
     */
    public void end(ExcelReader e){

	length++;
	e.getStrValue();
	checkTheCell = e.getStrValue();

    }

    /**This method dial with normal or batch choice
     */
    public void number(ExcelReader e){
	
	checkTheCell = e.getStrValue();
	if(checkTheCell.equals(sec)) second(e);
	else if(checkTheCell.equals(min))minute(e);
	else{ 
	    if(StartESFrame.verbose)
	    System.out.println("Time is not expressed in minutes or seconds. Check the worksheet");
	    System.exit(1);
	}

    }


    public void second(ExcelReader e){ 
	int numberOfSteps = 0;

	numberOfSteps = e.getIntValue();
	if(numberOfSteps == 0){length ++;
	checkTheCell = e.getStrValue();}
	else{
	    checkTheCell = e.getStrValue();
	    
	    if(checkTheCell.equals(slash) || checkTheCell.equals(backslash)){
		
		length +=4;
		e.getStrValue();
		checkTheCell = e.getStrValue();

	    }

	    else {
		length += numberOfSteps;
	    }

	}
    }

   public void minute(ExcelReader e){ 
      int numberOfSteps = 0;

      numberOfSteps = (e.getIntValue() * 60);
	    if(numberOfSteps == 0){length ++;
	    checkTheCell = e.getStrValue();}
	    else{
		checkTheCell = e.getStrValue();
	    
		if(checkTheCell.equals(slash) || checkTheCell.equals(backslash)){
		
		    length +=4;
		    e.getStrValue();
		    checkTheCell = e.getStrValue();

		}

		else {
		    length += numberOfSteps;
		}

	    }
	   
   }

    /**This is the method containing the iterator needed to launch the daily production of recipes. 
    * It take a look at the orderSequence arrays to determine which recipes must be done and how many times.
    * A request for which unit can do the first production phase of each recipe will be done to units or endUnits. 
    */
    public void distill(){
	//local counters
	int i, ii, iii, k, code, quantity, layerNumber;
	boolean recipeCodeNotFound;
	i=0;
	
	readOrderSequence();  //See below
	
	while(i < maxOrderSequence && orderSequence1[i] != 0){
	    code =  getOrderSequence1(i);
	    quantity = getOrderSequence2(i);
	    layerNumber = getOrderSequence3(i);

	    recipeCodeNotFound = true;
	    for (ii = 0; ii < recipeList.getCount() && recipeCodeNotFound; ii++)
		{
		    aRecipe = (Recipe) recipeList.atOffset(ii);
		    if(aRecipe.getCodeNumber() == code)recipeCodeNotFound = false;
		}
	    for(k = 0; k < quantity; k++){ 
		
		orderCount++;
		// creating an order
		anOrder = new Order(getZone(), orderCount,
				    Globals.env.getCurrentTime(),
				    aRecipe.getLength(), aRecipe.getOrderRecipe(), 
				    eSFrameModelSwarm, endUnitList);
		
		anOrder.setRecipeName(aRecipe.getRecipeName());

		//lm		
		//************************************************************
		// setting the layer (from 0 to totalLayerNumber-1)
		//************************************************************
		
		anOrder.setOrderLayer(layerNumber);
		if(StartESFrame.verbose)
		System.out.println("Order #" + anOrder.getOrderNumber() + 
			" is generated with Name: " + anOrder.getRecipeName() +
			" and layerNumber #" + anOrder.getOrderLayer());
		
		
		// add the active orders to the general order list (they will be
		// eliminated when dropped in a unit, being finished); this
		// list has been introduced for accounting purposes [may be it would
		// be better substitute it with a get to the units to know their
		// waiting lists] 
		orderList.addLast(anOrder);
				
		/**
		 * sending the order to the first production unit
		 * (we are acting as the Front End of the VE)
		 */
		assigningTool.assign(anOrder);
				
	    }
	    
	    i++;	
	}
	
	
    }
    
    /**This method reads from the worksheet containing, shift by shift, the sequence of orders to be launched and fills
     * in the orderSequence1 with the ID codes of recipes and the orderSequence2 with the quantities of each recipe.
     */
     
    public void readOrderSequence(){
	//local variables
	
	// lm 
	//*****************************************************
	// boolean firstTime has been moved to global variables
	//*****************************************************
	int i, numberOfShift;
	
	if(StartESFrame.verbose)
	System.out.println("firstTime is " + firstTime);
		
	i = 0;
	
	
	if(! worksheetOrderSequenceFileOpen && firstTime){ 
	    orderSequenceWorksheet = new ExcelReader("recipeData/orderStartingSequence.xls");
	    worksheetOrderSequenceFileOpen = true;
	    firstTime = false;
	    if(StartESFrame.verbose)
	    System.out.println("orderStartingSequence.xls has been open");
	}
	
	else if(! worksheetOrderSequenceFileOpen){ 
	    orderSequenceWorksheet = new ExcelReader("recipeData/orderSequence.xls");
	    worksheetOrderSequenceFileOpen = true;
	//lm
	    if(StartESFrame.verbose)
	    System.out.println("orderSequence.xls has been open");
	
	}
	
	//lm
	//****************************************************************
	// Read the shift number, added some prints for debugging purpose 
	//****************************************************************
	
	numberOfShift = orderSequenceWorksheet.getIntValue(); 
	if(StartESFrame.verbose)
	    System.out.println("The shift #" + numberOfShift + " has begun ");


	//lm
	//****************************************************************
	// A new bug. Zeroing orderSequenceX vectors
	//****************************************************************	
	for( int ii = 0; ii < maxOrderSequence; ii++){
	    orderSequence1[ii] = 0;
	    orderSequence2[ii] = 0;
	    orderSequence3[ii] = 0;
	    
	}	
	while(! orderSequenceWorksheet.eol()){
	    
	    //lm
	    //****************************************************
	    //Our new method used to check if there is a new layer
	    //****************************************************
	    checkForLayer(orderSequenceWorksheet);	
	    
	    orderSequence1[i] = errorIsNotAnInteger(orderSequenceWorksheet);
	   

	
	    orderSequence3[i] = currentLayer;
	    
	    errorIsNotAString(orderSequenceWorksheet);
	    orderSequenceWorksheet.getStrValue();
	    
	    orderSequence2[i] = errorIsNotAnInteger(orderSequenceWorksheet);
	    	   	
	   	//lm
	   	//**********************************
	   	// Some prints for debugging purpose 
	   	//**********************************
	    if(StartESFrame.verbose)
	    System.out.println("The recipe #" + orderSequence1[i] + " with quantity " + orderSequence2[i] 
	    	+ " and Layer " + orderSequence3[i] +" is starting production ");
	    
	    i++;    
	}
	
	errorIsNotAString(orderSequenceWorksheet);
	orderSequenceWorksheet.getStrValue();
	


	if(orderSequenceWorksheet.eof()){
	    //  System.out.println("The simulation finishes here. It is not the right and normal way, but is only for explanation. Thank you for your attention and interest.");
	    // System.exit(1);
	    worksheetOrderSequenceFileOpen = false;
	} 
    }    


	/**This method is used to check the corrispondence with the dictionary of production phases 
     */
    public int checkTheExistence(int c){
	int i, check;
	boolean dictionaryVoice = false;
	
	check = c;
	
	for(i = 0; i < dictionaryLength; i++){
	    if(check == dictionary[i])dictionaryVoice = true;
	} 
	
	if(check > 1000000000){ 
	    System.out.println("The value found is not a valid number because it is greater than a billion");
	    System.exit(1);
	}

	if(dictionaryVoice == false){ 
	    System.out.println("The value found is not a valid number of production");
	    System.exit(1);
	}
	
	return check;
    }
    
	//lm
	// ***************
	// Our new method
	// ***************
    
    /**This method is used to check the presence of a new layer
     */
    public void  checkForLayer(ExcelReader e){
	
	if(e.checkForLabelCell()){
		checkTheCell = e.getStrValue();
	   
		if(checkTheCell.equals(layer))
			currentLayer = e.getIntValue();
			
		//lm
		// **********************************
		// Some prints for debugging purpose 
		// **********************************
		if(StartESFrame.verbose)
		System.out.println("currentLayer now is " + currentLayer);
		
		}
    }


    /**This method is used to check if there are comments in the cells
     */
    public void checkForComments(String cTC){
	
	checkTheCell = cTC;
	while(checkTheCell.equals(gate)){
	    row++;
	    if(StartESFrame.verbose)
		System.out.println("The row # " + row + " contains a comment");

	    recipeWorksheet.getStrValue();
	    checkTheCell = recipeWorksheet.getStrValue();
	    if(! checkTheCell.equals(semicolon))errorIsNotAString(recipeWorksheet); 
	    checkTheCell = recipeWorksheet.getStrValue();
	}
    }



    /**This method is used to check if a type error occurs
     */
    public int errorIsNotAnInteger(ExcelReader e){
	
	if(e.checkForLabelCell()) 
	 {
	     System.out.println("The cell should contain an Integer Value");
	     System.exit(1);
	 }
	return e.getIntValue();
    }
    
    /**This method is used to check if a type error occurs
     */
    public void  errorIsNotAString(ExcelReader e){
	
	if(e.checkForLabelCell());
	else {
	    System.out.println("The cell should contain a String Value");
	    System.exit(1);
	}
    }

    /**This method is used to obtain the elements of the orderSequence1 array
     */
    public int getOrderSequence1(int j){
	
	return orderSequence1[j];
	
    }
    
    /**This method is used to obtain the elements of the orderSequence2 array
     */
    public int getOrderSequence2(int j){
	
	return orderSequence2[j];
	
    }
    
    
    /**This method is used to obtain the elements of the orderSequence3 array
     */
    public int getOrderSequence3(int j){
	
	return orderSequence3[j];
	
    }

} // OrderDistiller
